from pathlib import Path

import h5py
import numpy as np

import pytest


@pytest.fixture
def hdf5_path(tmp_path):
    path = Path(tmp_path / "data.hdf5")
    with h5py.File(path, "w") as f:
        g1 = f.create_group("g1")
        g1_1 = g1.create_group("g1_1")

        g2 = f.create_group("g2")

        f.create_dataset("ds", data=np.ones((5, 8)))
        g1.create_dataset("ds1", data=np.ones((5, 5)))
        g1_1.create_dataset("ds1_1", data=np.ones((10, 15)) * 2)
        g2.create_dataset("ds2", data=np.ones((3, 4, 5)) * 4)

    return path



#!/usr/bin/env python
import h5py
import numpy as np


with h5py.File("data.hdf5", "w") as f:
    g1 = f.create_group("g1")
    g1_1 = g1.create_group("g1_1")

    g2 = f.create_group("g2")

    ds = f.create_dataset("ds", data=np.ones((5, 8)))
    ds1 = g1.create_dataset("ds1", data=np.ones((5, 5)))
    ds1_1 = g1_1.create_dataset("ds1_1", data=np.ones((10, 15)) * 2)
    ds2 = g2.create_dataset("ds2", data=np.ones((3, 4, 5)) * 4)



from .version import version as __version__  # noqa: F401
from .commands import Command, format_dataset, format_obj, get_children, all_commands

__all__ = ["Command", "format_dataset", "format_obj", "get_children", "all_commands"]



from pathlib import PurePosixPath as H5Path
from typing import List, Union, Iterable
from enum import Enum, auto

from h5py import File, Group, Dataset

from prompt_toolkit.completion import Completion, Completer

ObjectType = Union[File, Group, Dataset]


def is_file(obj):
    return isinstance(obj, File)


def is_group(obj):
    return isinstance(obj, Group)


def is_dataset(obj):
    return isinstance(obj, Dataset)


def normalise_path(
    path: H5Path, current_gpath: H5Path = None, relative=False
) -> H5Path:
    path = H5Path(path)
    if current_gpath is not None:
        current_gpath = H5Path(current_gpath)
        path = current_gpath / path

    if not path.is_absolute():
        raise ValueError("Path must be absolute")

    parts: List[str] = []
    for part in path.parts:
        if set(part) == {"."}:
            for _ in part[1:]:
                try:
                    parts.pop()
                    if len(parts) == 0:
                        raise IndexError()
                except IndexError:
                    raise ValueError("Tried to traverse beyond root")
        else:
            parts.append(part)

    out = H5Path(*parts)
    if relative and current_gpath:
        out = out.relative_to(current_gpath)
    return out


def obj_name(obj: ObjectType) -> str:
    stripped = obj.name.strip("/")
    if not stripped:
        return "/"

    return stripped.split("/")[-1]


# TODO: consider replacing with OS signals
class Signal(Enum):
    SUCCESS = auto()
    FAILURE = auto()
    QUIT = auto()


class H5PathCompleter(Completer):
    """
    Complete for Path variables.
    :param get_paths: Callable which returns a list of directories to look into
                      when the user enters a relative path.
    :param file_filter: Callable which takes a filename and returns whether
                        this file should show up in the completion. ``None``
                        when no filtering has to be done.
    :param min_input_len: Don't do autocompletion when the input string is shorter.
    """

    def __init__(
        self,
        context,
        include_groups: bool = True,
        include_datasets: bool = True,
    ) -> None:
        self.context = context
        self.include_groups = include_groups
        self.include_datasets = include_datasets

    @property
    def gpath(self):
        return self.context.gpath

    @property
    def group(self):
        return self.context.group

    def get_completions(self, document, complete_event) -> Iterable[Completion]:
        text = document.text_before_cursor

        curr_path = H5Path(text)
        if text.endswith("/"):
            parent = curr_path
            prefix = ""
        else:
            parent = curr_path.parent
            prefix = curr_path.name

        try:
            parent_obj = self.group[str(normalise_path(parent, self.gpath))]
        except KeyError:
            return

        for name, obj in parent_obj.items():
            if not name.startswith(prefix):
                continue

            if (self.include_groups and is_group(obj)) or (
                self.include_datasets and is_dataset(obj)
            ):
                yield Completion(name[len(prefix) :], 0, display=name)



#!/usr/bin/env python
from setuptools import setup, find_packages
from pathlib import Path

here = Path(__file__).resolve().parent

long_description = (here / "README.md").read_text()

install_requires = ["h5py", "prompt-toolkit", "tree-format"]

setup(
    name="hcl",
    description="Interactive CLI for exploring HDF5 files",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/clbarnes/hcl",
    author="Chris L. Barnes",
    author_email="chrislloydbarnes@gmail.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    packages=find_packages(exclude="tests"),
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    python_requires=">=3.8, <4",
    install_requires=install_requires,
    entry_points={"console_scripts": ["hcl=hcl.__main__:main"]},
)



from typing import Sequence, Type
import shlex
from pathlib import Path
import logging
from functools import wraps
import sys

from h5py import File
from prompt_toolkit import print_formatted_text, PromptSession
from prompt_toolkit.completion import NestedCompleter
from prompt_toolkit.formatted_text import to_formatted_text

from .utils import H5Path, normalise_path, Signal, is_dataset, is_group
from .commands import Command


logger = logging.getLogger(__name__)


class Cli:
    def __init__(
        self,
        fpath,
        gpath="/",
        commands: Sequence[Type[Command]] = (),
        session_kwargs=None,
        print_kwargs=None,
        mode="r",
        interactive=True,
    ):
        self.fpath = Path(fpath) if fpath else None
        if not gpath:
            gpath = "/"
        elif not gpath.startswith("/"):
            gpath = "/" + gpath
        self.gpath = H5Path(gpath)
        self.print_kwargs = print_kwargs or dict()

        self.commands = dict()
        completers = dict()
        for cmd_cls in commands:
            c = cmd_cls(self)
            self.commands[c.name()] = c
            completers[c.name()] = c.completer()

        self.session_kwargs = {"completer": NestedCompleter(completers)}
        self.session_kwargs.update(session_kwargs or dict())
        self.mode = mode
        self.interactive = interactive

        self.session = None
        self.file = None
        self.group = None

    def __enter__(self):
        if self.fpath:
            self.file = File(self.fpath, mode=self.mode)
            self.group = self.file[str(self.gpath)]
            if not is_group(self.group):
                raise ValueError(f"Not a group: {self.H5Path}")
        return self

    def __exit__(self, exc_type, value, traceback):
        if self.file:
            self.file.close()
        self.file = None
        self.group = None

    def run_line(self, s):
        out = Signal.SUCCESS
        for cmd in s.split(";"):
            argv = shlex.split(cmd.strip())
            out = self.run_command(argv)
            if out != Signal.SUCCESS:
                break
        return out

    def run_command(self, argv):
        if not argv:
            return
        cmd, *args = argv

        try:
            fn = self.commands[cmd]
        except KeyError:
            self.print(f"Not a known command: {cmd}", file=sys.stderr)
            return Signal.FAILURE

        return fn(argv[1:])

    def run(self):
        prefix = f"{self.fpath}:{{}} $ "
        if not self.session:
            self.session = PromptSession(**self.session_kwargs)

        while True:
            line = self.session.prompt(prefix.format(self.gpath))
            result = self.run_line(line)
            if result == Signal.QUIT:
                break

    def change_group(self, path: H5Path):
        if self.file is None:
            raise RuntimeError("File not open")
        new_path = normalise_path(path, self.gpath)
        new_obj = self.file[str(new_path)]
        if is_dataset(new_obj):
            raise ValueError(f"Object at path is a dataset: {new_path}")
        self.gpath = new_path
        self.group = self.file[str(new_path)]

    @wraps(print_formatted_text)
    def print(self, *args, **kwargs):
        kwargs = {**self.print_kwargs, **kwargs}
        if kwargs.get("file", sys.stdout).isatty():
            print_formatted_text(*args, **kwargs)
        else:
            keep = {"sep", "end", "file", "flush"}
            if self.interactive:
                formatted_items = [tup[1] for tup in to_formatted_text(args)]
            else:
                formatted_items = args
            print(
                *formatted_items,
                **{k: v for k, v in kwargs.items() if k in keep},
            )



"""
CLI for interactive exploration of HDF5 files.
"""
from argparse import ArgumentParser
import logging
import shlex
from importlib import import_module
import sys
import os
from runpy import run_path

from .cli import Cli
from .commands import Command, all_commands as COMMANDS
from .utils import Signal
from .version import version as __version__

logger = logging.getLogger(__name__)


def is_py_file(s):
    return (os.sep in s or "/" in s or os.endswith(".py")) and os.path.exists(s)


def get_plugin_commands(import_path):
    mod_name, obj_name = import_path.split(":")
    if is_py_file(mod_name):
        obj = run_path(mod_name)[obj_name]
    else:
        mod = import_module(mod_name)
        obj = getattr(mod, obj_name)

    if hasattr(obj, "__call__"):
        obj = obj()

    if issubclass(obj, Command):
        out = [obj]
    else:
        out = list(obj)

    logger.debug(
        "Got commands from '%s' : %s", import_path, [f"'{c.__name__}'" for c in out]
    )
    return out


def main():
    parser = ArgumentParser(description=__doc__)
    parser.add_argument(
        "file",
        nargs="?",
        help=(
            "HDF5 file to explore. Add ':/path/to/group' to start in a specific group. "
            "If this is not given, only `--version`, `--help`, or "
            "`--command '<some_command> --help'` can be used."
        ),
    )
    parser.add_argument(
        "-c",
        "--command",
        help="Run a single command and exit.",
    )
    parser.add_argument(
        "-p",
        "--plugin",
        action="append",
        help=(
            "Import path for additional commands, in the form '{module}:{object}', "
            "where {module} can be an absolute import path, "
            "or the path to a python file which can be run; "
            "{object} can be a Command subclass, an iterable of them, "
            "or a callable returning either. "
            "Can be used multiple times."
        ),
        default=(),
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="Increase logging verbosity, up to -vv for debug.",
    )
    parser.add_argument(
        "--mode",
        "-m",
        default="r",
        help=(
            "Mode in which to open the file. "
            "'r' (default): Readonly, file must exist. "
            "'r+': Read/write, file must exist. "
            "'w': Create file, truncate if exists. "
            "'w-' or 'x': Create file, fail if exists. "
            "'a': Read/write if exists, create otherwise."
        ),
    )
    parser.add_argument(
        "--version", "-V", action="store_true", help="Print version and exit."
    )
    args = parser.parse_args()

    log_level = {
        0: logging.WARN,
        1: logging.INFO,
        2: logging.DEBUG,
    }.get(args.verbose, logging.DEBUG)

    logging.basicConfig(level=log_level)

    if args.version:
        print(__version__)
        sys.exit()

    piped = not sys.stdout.isatty() and bool(args.command)

    if not args.file:
        retval = 0
        if (
            args.command
            and ";" not in args.command
            and (
                args.command == "help"
                or args.command.startswith("help ")
                or "--help" in args.command
            )
        ):
            with Cli(None, commands=COMMANDS, interactive=not piped) as cli:
                result = cli.run_command(shlex.split(args.command))
            if result != Signal.SUCCESS:
                retval = 1
        else:
            logger.warning(
                "No file given and other args not interpreted as help message command"
            )
            parser.print_help()
            retval = 1
        sys.exit(retval)

    fpath_gpath = args.file.split(":")
    els = len(fpath_gpath)
    if els == 1:
        fpath = fpath_gpath[0]
        gpath = "/"
    elif els == 2:
        fpath, gpath = fpath_gpath
    else:
        raise ValueError(f"Got more than one group path from argument '{args.file}'")

    for import_path in args.plugin:
        COMMANDS.extend(get_plugin_commands(import_path))

    piped = not sys.stdout.isatty() and bool(args.command)

    with Cli(fpath, gpath=gpath, commands=COMMANDS, interactive=not piped) as cli:
        if args.command:
            cli.run_command(shlex.split(args.command))
        else:
            try:
                cli.run()
            except EOFError:
                cli.commands["exit"]()


if __name__ == "__main__":
    main()



from typing import Optional
import sys
from abc import ABC, abstractmethod
from argparse import ArgumentParser, Namespace
import logging
import pprint
from textwrap import indent

from h5py import Dataset
from tree_format import format_tree

from prompt_toolkit.completion import Completer, ThreadedCompleter
from .utils import H5Path, normalise_path, obj_name, Signal, H5PathCompleter, is_dataset


class Command(ABC):
    def __init__(self, context):
        self.context = context
        self.logger = logging.getLogger(f"{__name__}.{self.name()}")

    @abstractmethod
    def argument_parser(self) -> ArgumentParser:
        pass

    def __call__(self, argv=()) -> Signal:
        self.logger.debug("Called with arguments %s", argv)
        parser = self.argument_parser()
        try:
            parsed = parser.parse_args(argv)
        except SystemExit:
            return Signal.SUCCESS
        self.logger.debug("Parsed arguments to %s", parsed)
        try:
            result = self.run(parsed)
            if not isinstance(result, Signal):
                result = Signal.SUCCESS
        except Exception as e:
            self.logger.exception(str(e))
            self.context.print(f"Uncaught {type(e).__name__}: {e}", file=sys.stderr)
            result = Signal.FAILURE
        return result

    @abstractmethod
    def run(self, parsed_args: Namespace) -> Signal:
        pass

    @abstractmethod
    def name(self) -> str:
        pass

    def completer(self) -> Optional[Completer]:
        return None


class Ls(Command):
    def name(self):
        return "ls"

    def argument_parser(self):
        parser = ArgumentParser(self.name(), description="List members of a group")
        parser.add_argument(
            "path", nargs="*", help="Paths to list the members of", type=H5Path
        )

        return parser

    def ls_object(self, path: H5Path):
        obj_path = normalise_path(path, self.context.gpath, True)
        self.logger.debug("Normalised path to %s", obj_path)

        obj = self.context.group[str(obj_path)]
        self.logger.debug("Listing item at %s", obj.name)
        if is_dataset(obj):
            return str(path)
        else:
            rows = sorted(obj.keys())
            return "  ".join(rows)

    def run(self, parsed_args):
        paths = parsed_args.path
        if not paths:
            paths = [self.context.gpath]

        if len(paths) == 1:
            out = self.ls_object(paths[0])
        else:
            out = "\n\n".join(f"{path}:\n{self.ls_object(path)}" for path in paths)

        self.context.print(out)
        return Signal.SUCCESS

    def completer(self):
        return ThreadedCompleter(H5PathCompleter(self.context, include_datasets=False))


class Pwd(Command):
    def name(self):
        return "pwd"

    def argument_parser(self):
        return ArgumentParser(self.name(), description="Get working group")

    def run(self, parsed_args):
        self.context.print(str(self.context.gpath))
        return Signal.SUCCESS


class Cd(Command):
    def name(self):
        return "cd"

    def argument_parser(self):
        parser = ArgumentParser(self.name(), description="Change working group")
        parser.add_argument(
            "path", nargs="?", help="Path to next working group", type=H5Path
        )
        return parser

    def run(self, parsed_args):
        path = parsed_args.path
        if not path:
            return
        try:
            self.context.change_group(path)
            sig = Signal.SUCCESS
        except KeyError as e:
            self.context.print(str(e), file=sys.stderr)
            sig = Signal.FAILURE
        except ValueError as e:
            self.context.print(str(e), file=sys.stderr)
            sig = Signal.FAILURE
        return sig

    def completer(self):
        return ThreadedCompleter(H5PathCompleter(self.context, include_datasets=False))


class Exit(Command):
    def name(self):
        return "exit"

    def argument_parser(self):
        return ArgumentParser(self.name(), description="Quit hcl")

    def run(self, parsed_args):
        self.logger.debug("Quitting")
        return Signal.QUIT


class Attrs(Command):
    def name(self):
        return "attrs"

    def argument_parser(self):
        parser = ArgumentParser(
            "attrs", description="List attributes or look at one attribute"
        )
        parser.add_argument(
            "path",
            nargs="?",
            type=H5Path,
            help="Path to object whose attributes to check",
        )
        parser.add_argument(
            "attr",
            nargs="*",
            help="Any number of attributes to check. "
            "If none are given, show the names of all available attributes. "
            "If one is given, show the value of that attribute. "
            "If several are given, show the values for all of those attributes, in recfile-like format.",
        )
        parser.add_argument(
            "-a",
            "--all",
            action="store_true",
            help="Show all attribute values, in recfile-like format",
        )
        return parser

    def run(self, parsed_args):
        if not parsed_args.path:
            obj = self.context.group
        else:
            obj = self.context.group[str(parsed_args.path)]

        keys = parsed_args.attr
        if len(keys) == 1:
            self.context.print(pprint.pformat(obj.attrs[parsed_args.attr]))
            return

        if len(keys) == 0:
            sorted_keys = sorted(obj.attrs.keys())
            if parsed_args.all:
                keys = sorted_keys
            else:
                self.context.print("\n".join(sorted_keys))
                return

        for k in keys:
            v = obj.attrs[k]
            out = []
            formatted = pprint.pformat(v, indent=2)
            n_lines = formatted.count("\n") + 1
            if n_lines == 1:
                out.append(f"{k}: {formatted}")
            else:
                out.append(f"{k}:\n{indent(formatted, '+ ')}")
            self.context.print(*out, sep="\n")
        return Signal.SUCCESS

    def completer(self):
        return ThreadedCompleter(H5PathCompleter(self.context))


class AttributePrint(Command):
    _name: str
    _pprint = False

    _include_groups = True
    _include_datasets = True

    def name(self):
        return self._name

    def argument_parser(self):
        parts = []
        if self._include_groups:
            parts.append("group")
        if self._include_datasets:
            parts.append("dataset")
        s = " or ".join(parts)

        parser = ArgumentParser(self._name, description=f"Get {s} {self._name}")
        parser.add_argument("path", type=H5Path, help=f"Path to {s}")
        return parser

    def _format(self, obj):
        if self._pprint:
            return pprint.pformat(obj)
        else:
            return str(obj)

    def run(self, parsed_args):
        obj = self.context.group[str(parsed_args.path)]
        try:
            attr = getattr(obj, self._name)
        except AttributeError as e:
            self.context.print(str(e), file=sys.stderr)
            return Signal.FAILURE

        if attr is not None:
            self.context.print(self._format(attr))
        return Signal.SUCCESS

    def completer(self):
        return ThreadedCompleter(
            H5PathCompleter(self.context, self._include_groups, self._include_datasets)
        )


class Filename(AttributePrint):
    _name = "filename"


class Mode(AttributePrint):
    _name = "mode"


class Driver(AttributePrint):
    _name = "driver"


class Libver(AttributePrint):
    _name = "libver"


class UserblockSize(AttributePrint):
    _name = "userblock_size"


class Name(AttributePrint):
    _name = "name"


class Shape(AttributePrint):
    _name = "shape"
    _pprint = True
    _include_groups = False


class Dtype(AttributePrint):
    _name = "dtype"
    _include_groups = False


class Size(AttributePrint):
    _name = "size"
    _include_groups = False


class Maxshape(AttributePrint):
    _name = "maxshape"
    _pprint = True
    _include_groups = False


class Chunks(AttributePrint):
    _name = "chunks"
    _pprint = True
    _include_groups = False


class Compression(AttributePrint):
    _name = "compression"
    _include_groups = False


class CompressionOpts(AttributePrint):
    _name = "compression_opts"
    _pprint = True
    _include_groups = False


class Scaleoffset(AttributePrint):
    _name = "scaleoffset"
    _include_groups = False


class Shuffle(AttributePrint):
    _name = "shuffle"
    _include_groups = False


class Fletcher32(AttributePrint):
    _name = "fletcher32"
    _include_groups = False


class Fillvalue(AttributePrint):
    _name = "fillvalue"
    _include_groups = False


class IsVirtual(AttributePrint):
    _name = "is_virtual"
    _include_groups = False


class Help(Command):
    def name(self):
        return "help"

    def argument_parser(self):
        parser = ArgumentParser(self.name(), description="List available commands")
        parser.add_argument(
            "-s", "--short", action="store_true", help="Just show the names"
        )
        parser.add_argument(
            "-l",
            "--long",
            action="store_true",
            help="Show the full help text for every command",
        )
        return parser

    def run(self, parsed_args):
        commands = sorted(self.context.commands)
        if parsed_args.long:
            self.context.print(
                *(
                    self.context.commands[c].argument_parser().format_help()
                    for c in commands
                ),
                sep="\n\n",
            )
            return Signal.SUCCESS

        if parsed_args.short:
            self.context.print(*commands, sep="\n")
            return Signal.SUCCESS

        indent = max(len(c) for c in commands) + 4
        out = []
        for c in commands:
            desc = self.context.commands[c].argument_parser().description
            short = desc.split(".")[0] + "."
            out.append(f"{c}{' ' * (indent - len(c))}{short}")

        self.context.print(*out, sep="\n")
        return Signal.SUCCESS


def format_shape(ds: Dataset) -> str:
    return "x".join(str(s) for s in ds.shape)


def format_dataset(ds: Dataset):
    try:
        dtype = str(ds.dtype)
    except Exception:
        dtype = "<UNKNOWN>"
    return f"{obj_name(ds)}\t{format_shape(ds)}\t{dtype}"


def format_obj(obj):
    if is_dataset(obj):
        return format_dataset(obj)
    name = obj_name(obj)
    return name


def get_children(obj):
    if is_dataset(obj):
        return []
    else:
        return [v for _, v in sorted(obj.items())]


class Tree(Command):
    def name(self):
        return "tree"

    def argument_parser(self):
        parser = ArgumentParser(self.name(), description="Show hierarchy as a tree")
        parser.add_argument("path", nargs="*", type=H5Path, help="Groups to show")
        return parser

    def run(self, parsed_args):
        if not parsed_args.path:
            objs = [self.context.group]
        else:
            objs = [self.context.group[str(p)] for p in parsed_args.path]
        out = [
            format_tree(obj, format_node=format_obj, get_children=get_children)
            for obj in objs
        ]
        self.context.print(*out, sep="\n\n")
        return Signal.SUCCESS

    def completer(self):
        return H5PathCompleter(self.context, include_datasets=False)


# class Cp(Command):
#     def name(self):
#         return "cp"

#     def argument_parser(self):
#         parser = ArgumentParser(
#             self.name(),
#             description="Copy an object from one path to another; always recursive.",
#         )
#         parser.add_argument(
#             "source",
#             nargs="+",
#             help="Original object to copy. "
#             "Trailing slash means copy all objects from the given group, not the group itself.",
#         )
#         parser.add_argument(
#             "dest",
#             help="Destination to copy to. "
#             "Trailing slash OR multiple sources means copy *into* the given group, keeping original names. "
#             "Otherwise, copy to new group/dataset of the given name.",
#         )
#         return parser

#     def run(self, parsed_args):
#         sources = []
#         for src in parsed_args.source:
#             if src.endswith("/") and len(src) > 1:
#                 for member in self.context.group[str(src[:-1])].values():
#                     sources.append(member)
#             else:
#                 sources.append(self.context.group[str(src)])

#         dest = parsed_args.dest
#         into_group = len(sources) > 1 or dest.endswith("/")
#         if len(dest) > 1:
#             dest = dest.rstrip("/")

#         if into_group:
#             for src in sources:
#                 self.context.group.copy(src, dest)
#         else:
#             src = sources.pop()
#             dest_path = H5Path(dest)
#             self.context.group.copy(src, str(dest_path.parent), dest_path.name)
#         return Signal.SUCCESS


all_commands = [
    Ls,
    Pwd,
    Cd,
    Exit,
    Attrs,
    Filename,
    Mode,
    Driver,
    Libver,
    UserblockSize,
    Name,
    Shape,
    Dtype,
    Size,
    Maxshape,
    Chunks,
    Compression,
    CompressionOpts,
    Scaleoffset,
    Shuffle,
    Fletcher32,
    Fillvalue,
    IsVirtual,
    Help,
    Tree,
    # Cp,
]
